from sqlalchemy.sql.expression import select, exists
from sqlalchemy.sql.dml import Insert as insert
from sqlalchemy import func, or_, create_engine
from sqlalchemy.orm import sessionmaker
import sqlalchemy as sa
import warnings

warnings.filterwarnings("ignore", category=sa.exc.SAWarning)

engine = create_engine('trino://user@localhost:8080/iceberg')
connection = engine.connect()
SessionLocal = sessionmaker(
    bind=engine,
    autocommit=False,
    autoflush=False,
)


def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


class BaseQuery:
    """Base class for database queries.Provides common database operations."""

    @classmethod
    def session_db(cls):
        """Get the database session."""    
        return next(get_db())

    @classmethod
    def select_all(cls, table, columns=None, order_by=None): #order tyoe
        """Select all rows from the specified table."""
        if columns is None:
            query = select(table)
        else:
            query = select([getattr(table.c, col) for col in columns])
        if order_by:
            query = query.order_by(table.c[order_by])
        return cls.session_db().execute(query).mappings().all()

    @classmethod
    def select_columns(cls, table, *columns):
        """Select specific columns from the specified table."""   
        query = select([getattr(table.c, col) for col in columns])
        return cls.session_db().execute(query).mappings().all()

    @classmethod
    def select_with_condition(cls, table, condition, columns=None, order_by=None):
        """Select rows from the specified table with a given condition."""
        if columns is None:
            query = select(table).where(condition)
        else:
            query = select([getattr(table.c, col) for col in columns]).where(condition)
        if order_by:
            query = query.order_by(table.c[order_by])
        return cls.session_db().execute(query).mappings().all()


    @classmethod
    def select_with_join(cls, table1, table2, on1, table3=None, on2=None, table4=None, on3=None, table5=None, on4=None,
                         condition=None, join_type='inner', columns=None, order_by=None):
        """Select rows from the specified tables with a given join condition."""
        #     join.full: " FULL OUTER JOIN "
        #     join.isouter: " LEFT OUTER JOIN "
        #     join: " INNER JOIN "
        
        # Determine the join type
        is_full_join = False
        is_outer_join = False
        if join_type == 'inner':
            pass
        elif join_type == 'full':
            is_full_join = True
        elif join_type == 'right':
            temp = table1
            table1 = table2
            table2 = temp
            is_outer_join = True
        elif join_type == 'left':
            is_outer_join = True

        columns = []
        tables_alias = [(table1, table1.name), (table2, table2.name), (table3, table3.name)] if table3 is not None else [(table1, table1.name), (table2, table2.name)]
        if table4 is not None:
            tables_alias.append((table4, table4.name))
        if table5 is not None:
            tables_alias.append((table5, table5.name))
        for table_alias, prefix in tables_alias:
            for column in table_alias.c:
                columns.append(column.label(f"{prefix}_{column.name}"))

        query = select(*columns).join(table2, on1, full=is_full_join, isouter=is_outer_join)
        if table3 is not None:
            query = query.join(table3, on2, full=is_full_join, isouter=is_outer_join)
        if table4 is not None:
            query = query.join(table4, on3, full=is_full_join, isouter=is_outer_join)
        if table5 is not None:
            query = query.join(table5, on4, full=is_full_join, isouter=is_outer_join)
        query=query.where(condition) if condition is not None else query
        if order_by is not None:
            query = query.order_by(order_by)
        return cls.session_db().execute(query).mappings().all()


    @classmethod
    def insert_data(cls, table, data, pk="id",  unique_constraints=[]):
        """    Insert data into the specified table."""
        try:
            not_unique = False
            if unique_constraints:
                unique_data = {key: data[key] for key in unique_constraints}
                not_unique = cls._unique_constraints(table=table, data=unique_data)
            if not(not_unique):
                data[pk] = cls._next_pkey(table,pk)
                data = cls._created_timestamp(table=table, data=data)
                query = insert(table).values(**data)
                cls.session_db().execute(query)
                return True
            else:
                print("Not Unique")
                return False
        except Exception as e:
            print(f"Exception occured: {e}")
            return False


    @classmethod
    def update_data(cls, table, condition, data,unique_constraints=[]):
        """    Update rows in the specified table with a given condition."""
        try: 
            not_unique = False
            if unique_constraints:
                unique_data = {key: data[key] for key in unique_constraints}
                not_unique = cls._unique_constraints(table=table, data=unique_data)
            if not(not_unique):
                data = cls._updated_timestamp(table=table, data=data)
                query = table.update().where(condition).values(**data)
                cls.session_db().execute(query)
                return True
            else:
                print("Not Unique")
                return False
        except Exception as e:
            print("Exception occured: {e}")
            return False

    @classmethod
    def delete_data(cls, table, condition):
        """    Delete rows from the specified table with a given condition.""" 
        query = table.delete().where(condition)
        cls.session_db().execute(query)

    @classmethod
    def delete_all(cls, table):
        """Delete all records from the specified table."""
        session = cls.session_db()
        session.query(table).delete()
        session.commit()

    @classmethod
    def count_rows(cls, table):
        """    Count the number of rows in the specified table.""" 
        query = select(func.count()).select_from(table)
        return cls.session_db().execute(query).scalar()

    @classmethod
    def exists(cls, table, condition):
        """Check if rows exist in the specified table with the given condition."""
        query = select(exists().where(condition))
        return cls.session_db().execute(query).scalar()

    @classmethod
    def _next_pkey(cls, table,pk="id"):
        """ Get the next primary key integer. """
        query = select(func.max(table.c[pk]))
        max_id = cls.session_db().execute(query).scalar()
        if max_id is None:
            return 1
        return max_id + 1
        
    @classmethod
    def _unique_constraints(cls, table, data):
        """Check if data exists in the specified table based on unique constraints."""
        conditions = [(getattr(table.c, column) == value) for column, value in data.items()]
        query = select(func.count()).where(or_(*conditions))
        result = cls.session_db().execute(query).scalar()
        return result > 0

    @classmethod
    def _created_timestamp(cls, table, data):
        if hasattr(table.c, 'created_date'):
            data['created_date'] = func.now()
        if hasattr(table.c, 'updated_date'):
            data['updated_date'] = func.now()
        if hasattr(table.c, 'created_by'):
            data['created_by'] = 'admin'
        if hasattr(table.c, 'updated_by'):
            data['updated_by'] = 'admin'
        return data
        
    @classmethod
    def _updated_timestamp(cls, table, data):
        if hasattr(table.c, 'updated_date'):
            data['updated_date'] = func.now()
        if hasattr(table.c, 'updated_by'):
            data['updated_by'] = 'admin'
        return data
